/*
    dltrace 0.5
    Copyright (C) 2005 Richard Johnson <rjohnson@idefense.com>

    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation; either version 2 of the License, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this program; if not, write to the Free Software Foundation, Inc., 59 Temple
    Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <fcntl.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include "dltrace.h"
#include "melf.h"
#include "libdasm.h"

#define ELF_PAGESTART(_v) ((_v) & ~(unsigned long)(PAGE_SIZE-1))
#define ELF_PAGEOFFSET(_v) ((_v) & (PAGE_SIZE-1))
#define ELF_PAGEALIGN(_v) (((_v) + PAGE_SIZE - 1) & ~(PAGE_SIZE - 1))

static uint8_t * dt_search_libpath(const uint8_t *name);



int dt_load_binary(proc_t *ps)
{
	MELF *melf;
	uint8_t *path = ps->argv[0];

	if(!(melf = melf_open(path, MELF_ACCESS_READ)))
	{
		perror("melf_open");
		return -1;
	}
	ps->melf = melf;
 	ps->e_entry = melf->header.e_entry;
	ps->symbols = dt_load_symbols(melf, ps);

	if(dt_load_interp(melf, ps) < 0)
	{
		vlog(stdout, "fatal error: could not load interpreter\n");
		return -1;
	}
	
	if(dt_load_dynamic(melf, ps, 0) < 0)
	{
		vlog(stdout, "fatal error: could not load dynamic information\n");
		return -1;
	}
/*
	if(options.strace)
		dt_load_syscalls(melf, ps);	
*/
	return 0;
}

/* do not load symbols which have an addr of 0 */
list_t * dt_load_symbols(MELF *melf, proc_t *ps)
{
	ELF_SPEC_HEADER *shdr;
	Elf32_Sym *sym;
	list_t *symbols;
	uint8_t  *ptr, *libname = NULL;
	uint32_t libname_len = 0;


	// create empty node at end of symbol list
	symbols = malloc(sizeof(list_t));
	if(!symbols)
	{
		perror("malloc");
		return NULL;
	}
	memset(symbols, 0, sizeof(list_t));

	// get short name for library 
	ptr = melf->image + strlen(melf->image);
	while(ptr > (uint8_t *)melf->image && *ptr != '/')
		ptr--;

	if(ptr > (uint8_t *)melf->image)
		libname = strdup(++ptr);

	// now walk forward looking for a period
	ptr = libname;
	while(ptr++ < (libname + strlen(libname)))
		if(*ptr == '.' || *ptr == '-')
			*ptr = 0;

	libname_len = strlen(libname);
	
	if(!(shdr = melf_sectionGetEnum(melf)))
	{
		perror("melf_sectionGetEnum");
		return NULL;
	}
	
	while((shdr = melf_sectionEnumNext(melf, shdr)))
	{
		uint32_t i = 0, type, symname_len;
		uint8_t  *symname = NULL;
	
		type = melf_sectionGetType(melf, shdr);

		if(!(type == SHT_SYMTAB || type == SHT_DYNSYM))
			continue;

		while((sym = melf_symbolTableEnum(melf, shdr, i++)) != NULL)
		{
			symbol_t *syminfo = malloc(sizeof(symbol_t));
			if(!syminfo)
			{
				perror("malloc");
				return NULL;
			}
			memset(syminfo, 0, sizeof(symbol_t));
		
			syminfo->size = melf_symbolGetSize(melf, shdr, sym);
			if(syminfo->size == 0)
			{
				free(syminfo);
				continue;
			}
			
			syminfo->addr = melf_symbolGetValue(melf, shdr, sym);
			if(syminfo->addr == 0)
			{
				free(syminfo);
				continue;
			}
			
			symname = (uint8_t *)melf_symbolGetName(melf, shdr, sym);
			if(symname == NULL)
			{
				free(syminfo);
				continue;
			}
			
			symname_len = strlen(symname);
			if(symname_len == 0)
			{
				free(syminfo);
				continue;
			}
			syminfo->name = malloc(symname_len + libname_len + 2);
			if(!syminfo->name)
			{
				free(syminfo);
				continue;
			}
			sprintf(syminfo->name, "%s!%s", libname, symname);
	
			list_push(&symbols, syminfo);
		}
		
		if(options.verbose > 2)
			vlog(stdout, "load_symbols: %-10s contained %d symbols\n",
				melf_sectionGetName(melf, shdr),
				(melf_sectionGetSize(melf, shdr) / melf_sectionGetEntrySize(melf, shdr)));
		
	}	
	return symbols;
}


int dt_load_interp(MELF *melf, proc_t *ps)
{
	int fd;
	uint8_t *buf = NULL, *interp = NULL;
	Elf32_Ehdr *ehdr;
	Elf32_Phdr *phdr;
	struct stat st; 
	
	if(stat(ps->argv[0], &st) < 0)
	{
		perror("stat");
		return -1; 
	}
	
	fd = open(ps->argv[0], O_RDONLY);
	if(fd < 0)
	{
		perror("fatal error: could not open interpreter");
		return -1;
	}
	

	buf = mmap(NULL, st.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
	if(buf == MAP_FAILED)
	{
		perror("mmap");
		return -1;
	}
	
	ehdr = (Elf32_Ehdr *) buf;
	phdr = (Elf32_Phdr *)((uint8_t *)ehdr + ehdr->e_phoff);
	while(phdr->p_type != PT_INTERP)
	{
		if(phdr > (Elf32_Phdr *)(ehdr + ehdr->e_phoff + (ehdr->e_phnum * ehdr->e_phentsize)))
			break;
		phdr = (Elf32_Phdr *)((uint8_t *)phdr + ehdr->e_phentsize);
	}
	if(phdr->p_type != PT_INTERP)
	{
		vlog(stderr, "fatal error: could not find program interpreter\n");
		return -1;
	}

	interp = strdup(buf + phdr->p_offset);
	munmap(buf, 0);
	close(fd);

	if(options.verbose)
	vlog(stdout, "dltrace: loading interpreter: %s\n", interp);
	
	if(dt_load_library(ps, interp, 0) < 0)
	{
		vlog(stderr, "fatal error: could not load program interpreter\n");
		return -1;
	}
	ps->interp = ps->liblist->data;
	
	if(munmap(buf, st.st_size) < 0)
		vlog(stderr, "warning: could not munmap() interp\n");

	return 0;
}



int dt_load_dynamic(MELF *melf, proc_t *ps, uint32_t depth)
{
	ELF_SPEC_HEADER *shdr, *strtab;
	Elf32_Dyn *dyn;
	int num_entries;
	int i;
	uint8_t *name = NULL;
	uint8_t *path = NULL;

	depth++;
	
	if(depth < 5 && (shdr = melf_sectionFindType(melf, SHT_DYNAMIC)) != NULL)
	{
		strtab = melf_sectionGetStringTableHeader(melf, shdr);
		num_entries = melf_sectionGetSize(melf, shdr) / melf_sectionGetEntrySize(melf, shdr);
		for(i = 0; i < num_entries; i++)
		{
			path = NULL;
			dyn = melf_dynamicGetIndex(melf, shdr, i);
			if(dyn->d_tag == DT_NEEDED)
			{
				name = (uint8_t *)melf_stringTableGetString(melf, strtab, dyn->d_un.d_val);
				if(name != NULL)
				{
					// skip libpthread until issues within libmelf are addressed
					if(strncmp(name, "libpthread", 10) == 0)
						continue;
					if(*name != '/')
						path = dt_search_libpath(name);
					else
						path = name;
				}
				if(path !=NULL)
				{
					dt_load_library(ps, path, depth);
				}
			}
		}
	depth--;
	}
	
	return 0;
}

#define MAX_PATHS 128
static uint8_t * dt_search_libpath(const uint8_t *name)
{
	uint8_t *p, *paths[MAX_PATHS] = { "/lib","/usr/lib" };
	int path_idx = 2;
	FILE *fp;
	int i;
	
	if((fp = fopen("/etc/ld.so.conf", "r")) != NULL)
	{
		uint8_t buf[1024];
		while(fgets(buf, sizeof(buf) - 1 , fp) != NULL)
		{
			p = malloc(strlen(buf) + 1);
			if(!p)
			{
				perror("malloc");
				return NULL;
			}
			memset(p, 0, strlen(buf));
			strncpy(p, buf, strlen(buf));
			p[strlen(p) - 1] = 0; // chop off newline
			paths[path_idx] = p;
			path_idx++;
			memset(buf, 0, 1024);
		}
	}
	
	for(i = 0; i < path_idx; i++)
	{
		uint8_t *path = malloc(1024);
		
		if(!path)
		{
			perror("malloc");
			return NULL;
		}
		memset(path, 0, 1024);

		snprintf(path, 1024, "%s/%s", paths[i], name);
		if (access(path, F_OK) == 0)		/* We found it */
		{
			uint8_t *rpath = malloc(1024);
			memset(rpath, 0, 1024);
			if(readlink(path, rpath, 1024) < 0)
			{
				free(rpath);
				return path;
			}
			else
			{
				memset(path, 0, 1024);
				snprintf(path, 1024, "%s/%s", paths[i], rpath);
				return path; // return symbolic link
			}
		}
		else
			free(path);
	}

	if(fp) 
		fclose(fp);
	return NULL;
}

uint32_t dt_elf_hash_file(uint8_t *path)
{
	int fd;
	uint8_t *buf;
	uint32_t hash;
	
	fd = open(path, O_RDONLY);
	if(fd < 0)
	{
		perror("open");
		return -1;
	}

	if(!(buf = mmap(0, HASH_DATASZ, PROT_READ, MAP_PRIVATE, fd, 0)))
	{	
		perror("mmap");
		return 0;
	}

	hash = dt_hash_calc(buf, HASH_DATASZ, HASH_KEY);
	munmap(buf, HASH_DATASZ);
	
	return hash;
}

int32_t  dt_load_library(proc_t *ps, uint8_t *path, uint32_t depth)
{
	MELF *melf;
	library_t *lib = malloc(sizeof(library_t));
	list_t *libs = (list_t *)ps->liblist;
	
	memset(lib, 0, sizeof(library_t));
	
	while(libs != NULL)
	{
		if(strncmp(((library_t *)libs->data)->path, path, strlen(path)) == 0)
			return 0;
		libs = libs->next;
	}
	
	if(options.verbose)
		vlog(stderr, "dltrace: loading library %s\n", path);

	if(!(melf = melf_open(path, MELF_ACCESS_READ)))
	{
		printf("melf_open %s", path);
		perror("melf_open");
		return -1;
	}
	lib->melf = melf;
	lib->path = strdup(path);
	lib->hash = dt_elf_hash_file(path);
	lib->symbols = dt_load_symbols(melf, ps);

	dt_load_dynamic(melf, ps, depth);	
	list_push(&ps->liblist, lib);
	
	return 0;
}


#if 0
#define MIN(x, y) ((x) < (y) ? (x) : (y))
int dt_disas(uint8_t *data, uint32_t size)
{
	INSTRUCTION inst;	// declare struct INSTRUCTION
	int i, c = 0, bytes = 0, format = FORMAT_INTEL, len;
	uint8_t string[256];


	while (c < size)
	{
		/*
		 * get_instruction() has the following parameters:
		 *
		 * - &inst is a pointer to struct INSTRUCTION
		 * - data + c is pointer to data to be disassembled
		 * - disassemble in 32-bit mode: MODE_32 
		 */
		len = get_instruction(&inst, data + c, MODE_32);
		if((inst.type == INSTRUCTION_TYPE_INT) && (inst.op1.immediate == 0x80))
		{
		// Illegal opcode or opcode longer than remaining buffer
			if (!len || (len + c > size)) {
			vlog(stdout, "%.8x  ", c);
				if (bytes) {
				vlog(stdout, "%.2x  ", data[c]);
					for (i = 1; i < bytes*2 - 1; i++)
					vlog(stdout, " ");
				}
				if (format == FORMAT_INTEL)
				vlog(stdout, "db 0x%.2x\n", *(data + c));
				else
				vlog(stdout, ".byte 0x%.2x\n", *(data + c));
				c++;
				continue;
			}

			/*
			 * Print absolute offset and raw data bytes up to 'bytes'
			 * (not needed, but looks nice).
			 *	
			 */
			vlog(stdout, "%.8x  ", c);
			if (bytes) {
				for (i = 0; i < MIN(bytes, len); i++)
				vlog(stdout, "%.2x", data[c + i]);
			vlog(stdout, "  ");
				for (i = MIN(bytes, len); i < bytes*2 - len; i++)
				vlog(stdout, " ");
			}
		
			/*
			 * Print the parsed instruction, format using user-supplied
			 * format. We could of course format the instruction in some
			 * other way by accessing struct INSTRUCTION members directly.
			 */
			get_instruction_string(&inst, format, (DWORD)c, string, sizeof(string));
			vlog(stdout, "%s\n", string);
			
			c += len;
		}
		else
		{
			c+=len;
			continue;
		}
	} 

	return 0;



}

int32_t dt_load_syscalls(proc_t *ps, uint8_t *path)
{
	MELF *melf;
	Elf32_Ehdr *ehdr;
	ELF_SPEC_HEADER *phdr;
	uint8_t *buf = NULL; 
	uint32_t fd, i, c, ret;
	struct stat st;


	ret = stat(path, &st);
	if(ret < 0)
	{
		perror("stat");
		return -1;
	}

	if(!(buf = dt_map_file(path, st.st_size)))
	{
		vlog(stderr, "dt_load_syscalls: could not map file for analysis\n");
		return -1;
	}

	melf = melf_open(path, MELF_ACCESS_READ);
	if(!melf)
	{
		perror("melf_open");
		return -1;
	}

	phdr = melf_programGetEnum(melf);
	if(phdr == NULL)
	{
		vlog(stderr, "dltrace syscalls: could not get phdr enum\n");
		return -1;
	}

	while(phdr)	
	{
		uint32_t offset = 0, size = 0;
		uint8_t  *load = NULL;
		
		if(!(melf_programGetType(melf, phdr) == PT_LOAD))
		{
			phdr = melf_programEnumNext(melf, phdr);
			continue;
		}
		size = melf_programGetVirtualSize(melf, phdr);
		offset = melf_programGetPhysicalAddress(melf, phdr);

	vlog(stdout, "PT_LOAD offset: %08x  size: %08x bytes\n", (unsigned int)offset, (unsigned int)size);
		load = buf + offset;
		dt_disas(load, size);
		phdr = melf_programEnumNext(melf, phdr);
	}

	munmap(buf, st.st_size);

	return 0;
	
}
#endif

